/**
 * Copyright (c) Facebook, Inc. and its affiliates.
 *
 * This source code is licensed under the MIT license found in the
 * LICENSE file in the root directory of this source tree.
 *
 *  strict-local
 * @format
 */

/* eslint-env worker, serviceworker */
"use strict";

function asyncGeneratorStep(gen, resolve, reject, _next, _throw, key, arg) {
  try {
    var info = gen[key](arg);
    var value = info.value;
  } catch (error) {
    reject(error);
    return;
  }
  if (info.done) {
    resolve(value);
  } else {
    Promise.resolve(value).then(_next, _throw);
  }
}

function _asyncToGenerator(fn) {
  return function() {
    var self = this,
      args = arguments;
    return new Promise(function(resolve, reject) {
      var gen = fn.apply(self, args);
      function _next(value) {
        asyncGeneratorStep(gen, resolve, reject, _next, _throw, "next", value);
      }
      function _throw(err) {
        asyncGeneratorStep(gen, resolve, reject, _next, _throw, "throw", err);
      }
      _next(undefined);
    });
  };
}

const WebSocketHMRClient = require("../WebSocketHMRClient");

const bundleCache = require("./bundleCache");

const bundleToString = require("./bundleToString");

const patchBundle = require("./patchBundle");

const stringToBundle = require("./stringToBundle");

function fetchBundle(_x) {
  return _fetchBundle.apply(this, arguments);
}

function _fetchBundle() {
  _fetchBundle = _asyncToGenerator(function*(bundleReq) {
    const bundleRes = yield fetch(bundleReq, {
      includeCredentials: true
    });
    return stringToBundle(yield bundleRes.text());
  });
  return _fetchBundle.apply(this, arguments);
}

function defaultGetHmrServerUrl(bundleReq, revisionId) {
  const bundleUrl = new URL(bundleReq.url);
  return `${bundleUrl.protocol === "https:" ? "wss" : "ws"}://${
    bundleUrl.host
  }/hot?revisionId=${revisionId}`;
}

function defaultGetDeltaBundle(_x2, _x3) {
  return _defaultGetDeltaBundle.apply(this, arguments);
}

function _defaultGetDeltaBundle() {
  _defaultGetDeltaBundle = _asyncToGenerator(function*(bundleReq, revisionId) {
    const url = new URL(bundleReq.url);
    url.pathname = url.pathname.replace(/\.(bundle|js)$/, ".delta");
    url.searchParams.append("revisionId", revisionId);
    const res = yield fetch(url.href, {
      includeCredentials: true
    });
    return yield res.json();
  });
  return _defaultGetDeltaBundle.apply(this, arguments);
}

function defaultOnUpdate(clientId, update) {
  clients.get(clientId).then(client => {
    if (client != null) {
      client.postMessage({
        type: "METRO_UPDATE",
        update
      });
    }
  });
}

function defaultOnUpdateStart(clientId) {
  clients.get(clientId).then(client => {
    if (client != null) {
      client.postMessage({
        type: "METRO_UPDATE_START"
      });
    }
  });
}

function defaultOnUpdateError(clientId, error) {
  clients.get(clientId).then(client => {
    if (client != null) {
      client.postMessage({
        type: "METRO_UPDATE_ERROR",
        error
      });
    }
  });
}

function createDeltaClient() {
  let _ref =
      arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : {},
    _ref$getHmrServerUrl = _ref.getHmrServerUrl,
    getHmrServerUrl =
      _ref$getHmrServerUrl === void 0
        ? defaultGetHmrServerUrl
        : _ref$getHmrServerUrl,
    _ref$getDeltaBundle = _ref.getDeltaBundle,
    getDeltaBundle =
      _ref$getDeltaBundle === void 0
        ? defaultGetDeltaBundle
        : _ref$getDeltaBundle,
    _ref$onUpdateStart = _ref.onUpdateStart,
    onUpdateStart =
      _ref$onUpdateStart === void 0 ? defaultOnUpdateStart : _ref$onUpdateStart,
    _ref$onUpdate = _ref.onUpdate,
    onUpdate = _ref$onUpdate === void 0 ? defaultOnUpdate : _ref$onUpdate,
    _ref$onUpdateError = _ref.onUpdateError,
    onUpdateError =
      _ref$onUpdateError === void 0 ? defaultOnUpdateError : _ref$onUpdateError;

  const clientsByRevId = new Map();
  return (
    /*#__PURE__*/
    (function() {
      var _ref2 = _asyncToGenerator(function*(event) {
        const clientId = event.clientId;
        const bundleReq = event.request;
        let bundle = yield bundleCache.getBundle(bundleReq);

        if (bundle == null) {
          // We couldn't retrieve a delta bundle from either the delta cache nor the
          // browser cache. This can happen when the browser cache is cleared but the
          // service worker survives. In this case, we retrieve the original bundle.
          bundle = yield fetchBundle(bundleReq);
        } else if (!__DEV__) {
          try {
            const delta = yield getDeltaBundle(bundleReq, bundle.revisionId);
            bundle = patchBundle(bundle, delta);
          } catch (error) {
            console.error("[SW] Error retrieving delta bundle", error);
            bundle = yield fetchBundle(bundleReq);
          }
        } else {
          const clientIds = clientsByRevId.get(bundle.revisionId);

          if (clientIds != null) {
            // There's already an update client running for this particular
            // revision id.
            clientIds.add(clientId);
          } else {
            const clientIds = new Set([clientId]);
            clientsByRevId.set(bundle.revisionId, clientIds);

            try {
              let currentBundle = bundle;
              bundle = yield new Promise((resolve, reject) => {
                let resolved = false;
                const wsClient = new WebSocketHMRClient(
                  getHmrServerUrl(bundleReq, currentBundle.revisionId)
                );
                wsClient.on("connection-error", error => {
                  reject(error);
                });
                wsClient.on("close", () => {
                  clientsByRevId.delete(currentBundle.revisionId);
                });
                wsClient.on("error", error => {
                  if (!resolved) {
                    reject(error);
                    return;
                  }

                  clientIds.forEach(clientId => onUpdateError(clientId, error));
                });
                wsClient.on("update-start", () => {
                  clientIds.forEach(clientId => onUpdateStart(clientId));
                });
                wsClient.on("update", update => {
                  if (resolved) {
                    // Only notify clients for later updates.
                    clientIds.forEach(clientId => onUpdate(clientId, update));
                  } // Transfers all clients to the new revision id.

                  clientsByRevId.delete(currentBundle.revisionId);
                  clientsByRevId.set(update.revisionId, clientIds);
                  currentBundle = patchBundle(currentBundle, {
                    base: false,
                    revisionId: update.revisionId,
                    modules: update.modules,
                    deleted: update.deleted
                  });
                  bundleCache.setBundle(bundleReq, currentBundle);

                  if (!resolved) {
                    resolved = true;
                    resolve(currentBundle);
                  }
                });
                wsClient.enable();
              });
            } catch (error) {
              console.error(
                "[SW] Error connecting to the update server. Try refreshing the page.",
                error
              );
              bundle = yield fetchBundle(bundleReq);
            }
          }
        }

        bundleCache.setBundle(bundleReq, bundle);
        const bundleString = bundleToString(bundle);
        const bundleStringRes = new Response(bundleString, {
          status: 200,
          statusText: "OK",
          headers: new Headers({
            "Cache-Control": "no-cache",
            "Content-Length": String(bundleString.length),
            "Content-Type": "application/javascript",
            Date: new Date().toUTCString()
          })
        });
        return bundleStringRes;
      });

      return function(_x4) {
        return _ref2.apply(this, arguments);
      };
    })()
  );
}

module.exports = createDeltaClient;
